#include <stdlib.h>
#include <windows.h>
#include <GL/gl.h>
#include <math.h>

#include "terrain.h"

#define RANDMASK RAND_MAX
#define MRANDOM (float)(rand() & (RANDMASK))/(RANDMASK)
#define SRANDOM ((MRANDOM) * 2) - 1

// draw lists
#define TERRAIN_LIST 1

#define WATERBOXSIZE 11
#define SAMPLESIZE 10
#define WATERSAMPLE 20

#define DEGTORAD				0.01745329

/*
 * spline(...)
 *
 *	interpolate a 1d spline
 *	taken from _Texturing & Modeling: A Procedural Approach_
 *		chapter 2 - by Darwyn Peachey
 */
float spline( float x, int nknots, float *knot )
{
	int span;
	int nspans = nknots - 3;
	float c0, c1, c2, c3;	// coefficients of the cubic

	if( nspans < 1)
	{
		// illegal
		return 0;
	}

	// find the appropriate 4-point span of the spline
	x = (x < 0 ? 0 : (x > 1 ? 1 : x)) * nspans;
	span = (int) x;
	if( span >= nknots -3 )
		span = nknots - 3;
	x -= span;
	knot += span;

	// evaluate the span cubic at x using horner's rule
	c3 = CR00*knot[0] + CR01*knot[1]
		+ CR02*knot[2] + CR03*knot[3];
	c2 = CR10*knot[0] + CR11*knot[1]
		+ CR12*knot[2] + CR13*knot[3];
	c1 = CR20*knot[0] + CR21*knot[1]
		+ CR22*knot[2] + CR23*knot[3];
	c0 = CR30*knot[0] + CR31*knot[1]
		+ CR32*knot[2] + CR33*knot[3];

	return ((c3*x + c2)*x + c1)*x + c0;
}

/*
 * initTerrain(..)
 *
 *	creates a zeroed buffer
 *	sets parameters and default values
 */
void initTerrain( terrain_t *ter, int xres, int zres, float xscale, float zscale, float scale )
{

	ter->hbuffer = (float *)calloc( sizeof(float) * xres * zres, 1 );
	ter->wbuffer = (float *)calloc( sizeof(float) * xres * zres, 1 );
	ter->xres = xres;
	ter->zres = xres;
	ter->xscale = xscale;
	ter->zscale = zscale;
	ter->scale = scale;

	if( ter->hbuffer == NULL || ter->wbuffer == NULL )
		MessageBox( NULL, "OUT OF MEMORY", "OUT OF MEMORY", MB_OK );
}

/*
 * freeTerrain(..)
 *
 *	deletes the buffers
 */
void freeTerrain( terrain_t *ter )
{
	int i;
	free( ter->hbuffer );
	free( ter->wbuffer );

	free( ter );
}

/*
 * createTerrain(...)
 *
 *	set remaining values
 *	use fracSynth Pass to generate height values
 *	create arrays for glDrawElements
 */
void createTerrain( terrain_t *ter, int seed, float fyscale, float freqscale, float initfreq, int numpass, int river, int twod )
{

	float freq;
	int pass;
	float yscale;
	float pt1[2], pt2[2];
	int i, j, k, cnt;

	// set values
	ter->seed = seed;
	ter->fyscale = fyscale;
	ter->freqscale = freqscale;
	ter->initfreq = initfreq;
	ter->numpass = numpass;
	srand( seed );

	// generate height values
	freq = ter->initfreq;
	pass = 0;
	yscale = ter->scale;
	for( pass=0;pass<numpass;pass++ )
	{
		fracSynthPass( ter->hbuffer, freq, yscale, ter->xres, ter->zres );
		freq += freqscale;
		yscale *= fyscale;
	}

	// add a river/stream

	if( river == 1)
	{
		pt1[0] = MRANDOM;
		pt1[1] = 0;
		pt2[0] = MRANDOM;
		pt2[1] = 1;
		generateRiver( ter->hbuffer, ter->wbuffer, ter->xres, ter->zres, pt1, pt2, 0 );
	}

	// find min and max
	ter->miny = ter->maxy = ter->hbuffer[0];
	for( i=0;i<ter->xres*ter->zres;i++ )
	{
		if( ter->hbuffer[i] < ter->miny )
			ter->miny = ter->hbuffer[i];
		else if( ter->hbuffer[i] > ter->maxy )
			ter->maxy = ter->hbuffer[i];
	}

	// create display list
	glNewList( TERRAIN_LIST, GL_COMPILE );
		for( i=0;i<ter->xres-1;i++ )
		{
			glBegin( GL_QUAD_STRIP );
				for( j=0;j<ter->zres;j++ )
				{
					float col = ((float)(ter->hbuffer[j*ter->xres+i+1] - ter->miny) / (ter->maxy - ter->miny )) * 0.4 + 0.3;
					glColor3f( col/4, col/3, col/5 );
					if( ter->wbuffer[j*ter->xres+i+1] > 0.4 )
						glColor3f( 0.0, (1-ter->wbuffer[j*ter->xres+i+1])/2, ter->wbuffer[j*ter->xres+i+1]/2 );
					glVertex3f( ((float)(i+1)/ter->xres * ter->xscale) - (ter->xscale/2),
						twod ? 0.0 : ter->hbuffer[j*ter->xres+(i+1)],
						((float)j/ter->xres * ter->zscale) - (ter->zscale/2) );
					col = ((float)(ter->hbuffer[j*ter->xres+i] - ter->miny) / (ter->maxy - ter->miny )) * 0.4 + 0.3;					glColor3f( col/4, col/3, col/4 );
					glColor3f( col/4, col/3, col/5 );
					if( ter->wbuffer[j*ter->xres+i] > 0.4 )
						glColor3f( 0.0, (1-ter->wbuffer[j*ter->xres+i])/2, ter->wbuffer[j*ter->xres+i]/2 );
					glVertex3f( ((float)i/ter->xres * ter->xscale) - (ter->xscale/2),
						twod ? 0.0 : ter->hbuffer[j*ter->xres+i],
						((float)j/ter->xres * ter->zscale) - (ter->zscale/2) );
				}
			glEnd();
		}
	glEndList();


}


/*
 * generate river(...)
 *
 * interpolate from pt1 to pt2 with some randomness
 * uses riverbox to create a reasonable looking river bed
 */
void generateRiver( float *hbuf, float *wbuf, int xres, int zres, float p1[2], float p2[2], float complexity )
{
	float riverbox[WATERBOXSIZE][WATERBOXSIZE];
	int i,j,k;
	float *xknots, *yknots, *zknots;
	float *points;
	float delta[3];
	float pt1[3];
	float pt2[3];
	int npts = complexity+4;
	int cnt;
	float ik, min;

	for( i=0;i<WATERBOXSIZE;i++ )
	{
		for( j=0;j<WATERBOXSIZE;j++ )
		{
			float fi=(float)(i-(WATERBOXSIZE/2))/(WATERBOXSIZE/2); float fj=(float)(j-(WATERBOXSIZE/2))/(WATERBOXSIZE/2);
			riverbox[i][j] = cos(fi) * cos(fj);
		}
	}

	// points (xyz for interpolation)
	points = (float*)calloc( sizeof(float)*npts*3, 1 );
	// knots for spline interpolation
	xknots = (float*)calloc( sizeof(float)*npts, 1 );
	yknots = (float*)calloc( sizeof(float)*npts, 1 );
	zknots = (float*)calloc( sizeof(float)*npts, 1 );

	// generate some points
	i = p1[0] * (xres-1);
	j = p1[1] * (zres-1);
	pt1[0] = i;
	pt1[1] = hbuf[j*xres+i];
	pt1[2] = j;
	i = p2[0] * (xres-1);
	j = p2[1] * (zres-1);
	pt2[0] = i;
	pt2[1] = hbuf[j*xres+i];
	pt2[2] = j;

	for( i=0;i<3;i++ )
		delta[i] = pt2[i]-pt1[i];

	// interpolate points
	for( i=0;i<npts;i++ )
	{
		for( j=0;j<3;j++ )
			points[i*3+j] = pt1[j] + ((float)(i-1))/(npts-3)*delta[j];
	}
	// add some non-linearity
	for( i=0;i<npts;i++ )
	{
		if( (i==1) || (i==npts-2) )
			continue;
		points[i*3+0] += SRANDOM/5.0*xres;
		points[i*3+1] += SRANDOM/5.0*xres;
	}
	points[3+0] += SRANDOM/5.0*xres;
	points[(npts-2)*3+0] += SRANDOM/5.0*xres;

	// spline interpolation - make the knots
	for( i=0;i<npts;i++ )
	{
		xknots[i] = points[i*3+0];
		yknots[i] = points[i*3+1];
		zknots[i] = points[i*3+2];
	}

	// total interpolated points
	cnt = 120/npts;
	ik = 0;
	free(points);
	points = (float*)calloc( sizeof(float)* cnt*npts*4,1 );

	for( j=0;j<=npts*cnt;j++ )
	{
		float d = ik/(npts*cnt);
		points[j*3+0] = spline( d, npts, xknots );
		points[j*3+1] = spline( d, npts, yknots );
		points[j*3+2] = spline( d, npts, zknots );
		ik += 1;

	}

	free( xknots );
	free( yknots );
	free( zknots );

	// displace terrain by water
	for( i=0;i<=npts*cnt;i++ )
	{
		float x,y,z;
		float wscale;
		int sx,sz;
		x = points[i*3+0];
		y = points[i*3+1];
		z = points[i*3+2];
		if( (x < 0) || (z < 0) )
			continue;
		sx = (int)(x);
		sz = (int)(z);
		if( i==npts*cnt )
			min = 1000;
		min = 500;
		// calculate minimum value of area
		for( j=0;j<SAMPLESIZE;j++ )
		{
			for( k=0;k<SAMPLESIZE;k++ )
			{
				if( (j-(SAMPLESIZE/2)+sz < 0) || (k-(SAMPLESIZE/2)+sx < 0) ||
					(j-(SAMPLESIZE/2)+sz >= zres) || (k-(SAMPLESIZE/2)+sx >= xres) )
					continue;
				if( hbuf[(j-(SAMPLESIZE/2)+sz)*xres+(k-(SAMPLESIZE/2)+sx)] < min )
					min = hbuf[(j-(SAMPLESIZE/2)+sz)*xres+(k-(SAMPLESIZE/2)+sx)];
			}
		}

		wscale = SRANDOM;
		// displace terrain
		for( j=0;j<WATERBOXSIZE;j++ )
		{
			for( k=0;k<WATERBOXSIZE;k++ )
			{
				if( (j-(WATERBOXSIZE/2)+sz < 0) || (k-(WATERBOXSIZE/2)+sx < 0) ||
					(j-(WATERBOXSIZE/2)+sz >= zres) || (k-(WATERBOXSIZE/2)+sx >= xres) )
					continue;
				hbuf[(j-(WATERBOXSIZE/2)+sz)*xres+(k-(WATERBOXSIZE/2)+sx)] -= (riverbox[j][k]/2);
				wbuf[(j-(WATERBOXSIZE/2)+sz)*xres+(k-(WATERBOXSIZE/2)+sx)] += riverbox[j][k]/cnt*2;
				if( wbuf[(j-(WATERBOXSIZE/2)+sz)*xres+(k-(WATERBOXSIZE/2)+sx)] )
					wbuf[0] = 1.0;
			}
		}
	}

	// free memory
	free( points );
}

/*
 * fracSynthPass(...)
 *
 *	generate basic points
 *	interpolate along spline & scale
 *	add to existing hbuffer
 */
void fracSynthPass( float *hbuf, float freq, float zscale, int xres, int zres )
{

	int i;
	int x, z;
	float *val;
	int max;
	float dfx, dfz;
	float *zknots, *knots;
	float xk, zk;
	float *hlist;
	float *buf;

	// how many to generate (need 4 extra for smooth 2d spline interpolation)
	max = freq + 2;
	// delta x and z - pixels per spline segment
	dfx = xres / (freq-1);
	dfz = zres / (freq-1);

	// the generated values - to be equally spread across buf
	val = (float*)calloc( sizeof(float)*max*max, 1 );
	// intermediately calculated spline knots (for 2d)
	zknots = (float*)calloc( sizeof(float)*max, 1 );
	// horizontal lines through knots
	hlist = (float*)calloc( sizeof(float)*max*xres, 1 );
	// local buffer - to be added to hbuf
	buf = (float*)calloc( sizeof(float)*xres*zres, 1 );

	// start at -dfx, -dfz - generate knots
	for( z=0; z<max; z++ )
	{
		for( x=0;x<max;x++ )
		{
			val[z*max+x] = SRANDOM;
		}
	}

	// interpolate horizontal lines through knots
	for( i=0;i<max;i++ )
	{
		knots = &val[i*max];
		xk = 0;
		for( x=0;x<xres;x++ )
		{
			hlist[i*xres+x] = spline( xk/dfx, 4, knots );
			xk += 1;
			if( xk >= dfx )
			{
				xk -= dfx;
				knots++;
			}
		}
	}

	// interpolate all vertical lines
	for( x=0;x<xres;x++ )
	{
		zk = 0;
		knots = zknots;
		// build knot list
		for( i=0;i<max;i++ )
		{
			knots[i] = hlist[i*xres+x];
		}
		for( z=0;z<zres;z++ )
		{
			buf[z*xres+x] = spline( zk/dfz, 4, knots ) * zscale;
			zk += 1;
			if( zk >= dfz )
			{
				zk -= dfz;
				knots++;
			}
		}
	}

	// update hbuf
	for( z=0;z<zres;z++ )
		for( x=0;x<xres;x++ )
			hbuf[z*xres+x] += buf[z*xres+x];

	free( val );
	free( buf );
	free( hlist );
	free( zknots );

}

void drawTerrainPoint( terrain_t *ter )
{
	int x, z;

	if( ter == NULL )
		return;

	glBegin( GL_POINTS );
		for( z=0;z<ter->zres;z++ )
		{
			for( x=0;x<ter->xres;x++ )
			{
				float col = ((ter->hbuffer[z*ter->xres+x] + 4) / 16) + 0.3;
				glColor3f( col, col, col );
				if( ter->wbuffer[z*ter->xres+x] > 0.7 )
					glColor3f( 0.0, 0.0, 1.0 );
				glVertex3f( ((float)x/ter->xres * ter->xscale) - (ter->xscale/2),
							ter->hbuffer[z*ter->xres+x],
							((float)z/ter->zres * ter->zscale) - (ter->zscale/2) );
			}
		}
	glEnd();

}

void drawTerrainPoly( terrain_t *ter )
{
	int i,j;

	if( ter == NULL )
		return;

	glColor3f( 0.4, 0.7, 0.4 );

	glCallList( TERRAIN_LIST );

}